#!/usr/bin/env python
# coding: utf-8

# # Experimentación TP2

# ### Setup

# Compilamos el código

# In[1]:


# get_ipython().system('cd .. && mkdir build')
# get_ipython().system('cd ../build/ && rm -rf *')
# get_ipython().system('cd ../build && cmake   -DPYTHON_EXECUTABLE="$(which python)"   -DCMAKE_BUILD_TYPE=Release ..')
# get_ipython().system('cd ../build && make install')


# Importamos módulos necesarios

# In[2]:


import metnum, os, csv, collections, numpy as np, pandas as pd, matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split, KFold
from sklearn.metrics import accuracy_score, cohen_kappa_score, precision_score, recall_score, f1_score
from time import time
from operator import add


# In[3]:


# Globals
colors = ["red", "orange", "blue", "green"]


# Importamos el CSV y lo guardamos en un dataframe

# In[4]:


df_base_completa_con_labels = pd.read_csv("../data/train.csv")

# Comenter esto para correr toda la base.
df_base_completa_con_labels = df_base_completa_con_labels[:5000]


# Definimos particiones de los datos (para train y test) y nos aseguramos que estén balanceadas.

# In[5]:


# seed = 5 # Nos da un ordenamiento balanceado
porcentaje_para_test = 0.2
KFOLDS = 5

labels_base_completa = df_base_completa_con_labels["label"].values.reshape(-1, 1)
df_base_completa = df_base_completa_con_labels[df_base_completa_con_labels.columns[1:]].values

# print(labels_base_completa)

# Genero K-folds
KFInstance = KFold(KFOLDS)
bases_test, bases_train, labels_train, labels_test = [], [], [], []
for train_index, test_index in KFInstance.split(df_base_completa):
    base_train, base_test = df_base_completa[train_index], df_base_completa[test_index]
    label_train, label_test = labels_base_completa[train_index], labels_base_completa[test_index]
    
    bases_train.append(base_train)
    bases_test.append(base_test) 
    labels_train.append(label_train) 
    labels_test.append(label_test)

# base_train, base_test, labels_train, labels_test = train_test_split(
#         df_base_completa, 
#         labels_base_completa, 
#         test_size = porcentaje_para_test,
#         random_state = seed
#     )

# Chequeamos que los datos estén balanceados.
tolerancia_de_balanceo = 0.1

for K in range(KFOLDS):
    total_labels = labels_train[K].shape[0]
    totales_por_digito = collections.Counter(labels_train[K].reshape(1,-1).tolist()[0])

    for i in range(0,9):
        porcentaje_digito_i = totales_por_digito[i] / total_labels
        if not abs(0.1-porcentaje_digito_i) < tolerancia_de_balanceo:
            raise NameError(f'Tolerancia inválida para el dígito {i}')


# In[6]:


# Función KNN
def runKNN(k, votation_method, base_train, labels_train, base_test):
    KNNClassifierInstance = metnum.KNNClassifier(k)
    KNNClassifierInstance.fit(base_train, labels_train)

    start_time = time()
    labels_test_predicted = KNNClassifierInstance.predict(base_test, votation_method)
    end_time = time()
    
    prediction_time = end_time - start_time
    
    return (labels_test_predicted, prediction_time)


# In[7]:


# Función para imprimir gráficos
def graficoTP2(filename, x, y, xlabel, ylabel, line_color = "black"):
#     plt.title(title)
#     plt.figure(name)
    plt.plot(x, y, marker = 'o', color = line_color)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.grid()
    plt.savefig(f'graficos/{filename}.png')
    # plt.show()


# In[8]:


def print_bold(text):
    print('\033[1m' + text + '\033[0m')


# ### Experimento 1
# Analizamos el mejor k y el mejor método de votación para kNN, la base de train seleccionada.

# In[9]:


MAX_K = 100
K_STEP = 5

ks = np.arange(0, MAX_K + 1, K_STEP)
votation_methods = ["uniform", "distance"]


# In[10]:


def Experimento1(base_train, base_test, labels_train, labels_test):
    accuracys = {} 
    f1s = {}
    times = {}

    for votation_method in votation_methods: 
        accuracys[votation_method] = []
        f1s[votation_method] = []
        times[votation_method] = []

    for k in ks:
        for votation_method in votation_methods: 
            labels_test_predicted, prediction_time = runKNN(k, votation_method, base_train, labels_train, base_test)
            
            accuracys[votation_method].append(accuracy_score(labels_test, labels_test_predicted))
            f1s[votation_method].append(f1_score(labels_test, labels_test_predicted, average = "macro"))
            times[votation_method].append(prediction_time)

        print(f'    k={k} de {MAX_K} procesado')
        
    return (accuracys, f1s, times)


# In[11]:


experimentPrefix = 'Experimento 1'

if not os.path.exists(f'resultados/{experimentPrefix}'):
    os.makedirs(f'resultados/{experimentPrefix}')
    
def Experimento1SaveResults(fold, cant_rows, accuracys, f1s, times):
    HEADERS = ["fold", "k", "votation_method", "accuracys", "f1s", "times"]
    with open(f'resultados/{experimentPrefix}/fold={fold}_raw_results.csv', mode='w') as results_file:
        results_writer = csv.writer(results_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        results_writer.writerow(HEADERS)
        for k in range(cant_rows):
            for votation_method in votation_methods:
                results_writer.writerow([
                    fold,
                    votation_method,
                    k * K_STEP, # Esto es medio feo acá pero es que estamos usando un range con step (refactorizar si hay tiempo)
                    accuracys[votation_method][k],
                    f1s[votation_method][k], 
                    times[votation_method][k]
                ])


# In[12]:


K_accuracys = []
K_f1s = []
K_times = []

import multiprocessing
print_bold(f'Corriendo Experimento 1 (buscar el mejor k y método de votación)')
with multiprocessing.Pool() as pool:
    lista_argumentos=zip(bases_train,bases_test,labels_train,labels_test)
    resultados = pool.starmap(Experimento1,lista_argumentos)
    #import pdb;pdb.set_trace()
    #accuracys = [l[0] for l in resultados]
    #f1s = [l[1] for l in resultados]
    #times = [l[2] for l in resultados]
#for K in range(KFOLDS):
    print(f'  Procesando fold = {K} ({KFOLDS} folds totales)')
###
    #accuracys, f1s, times = Experimento1(bases_train[K], bases_test[K], labels_train[K], labels_test[K])
    #import pdb;pdb.set_trace()
    #cant_rows=21
    #cant_rows = len(times['uniform'])
    for K in range(KFOLDS):
        accuracys = resultados[K][0]
        f1s = resultados[K][1]
        times = resultados[K][2]
        cant_rows = len(times['uniform'])
    #    import pdb;pdb.set_trace()
        Experimento1SaveResults(K, cant_rows, accuracys, f1s, times)
        K_accuracys.append(accuracys)
        K_f1s.append(f1s)
        K_times.append(times)


# In[13]:


# Calculo promedio de indicadores, generalizar a K folds si hay tiempo :( 
accuracys_average = {}
f1s_average = {}
times_average = {}

for votation_method in votation_methods:
    accuracys_average[votation_method] = np.average(
        np.array(
            [K_accuracys[0][votation_method], K_accuracys[1][votation_method], K_accuracys[2][votation_method],
            K_accuracys[3][votation_method], K_accuracys[4][votation_method]]
        ), axis=0)
    
    f1s_average[votation_method] = np.average(
        np.array(
            [K_f1s[0][votation_method], K_f1s[1][votation_method], K_f1s[2][votation_method],
            K_f1s[3][votation_method], K_f1s[4][votation_method]]
        ), axis=0)

    times_average[votation_method] = np.average(
        np.array(
            [K_times[0][votation_method], K_times[1][votation_method], K_times[2][votation_method],
            K_times[3][votation_method], K_times[4][votation_method]]
        ), axis=0)


# Nos quedamos con los mejores valores, los printeamos y guardamos

# In[14]:


print_bold('Resultados del Experimento 1')
experimentPrefix = 'Experimento 1'

HEADERS = [
    "fold", 
    "votation_method", 
    "best_accuracy",
    "best_k_for_accuracy",
    "best_f1",
    "best_k_for_f1",
    "time",
    "best_k_for_time"
]

if not os.path.exists(f'resultados/{experimentPrefix}'):
    os.makedirs(f'resultados/{experimentPrefix}')

with open(f'resultados/{experimentPrefix}/best_results.csv', mode='w') as results_file:
    results_writer = csv.writer(results_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
    results_writer.writerow(HEADERS)

    for K in range(KFOLDS):
        print(f'  Para fold = {K}')

        for votation_method in votation_methods:
            print(f'    método = {votation_method}')

            accuracys = K_accuracys[K]
            best_accuracy = max(accuracys[votation_method])
            best_k_for_accuracy = accuracys[votation_method].index(best_accuracy) * K_STEP

            f1s = K_f1s[K]
            best_f1 = max(f1s[votation_method])
            best_k_for_f1 = f1s[votation_method].index(best_f1) * K_STEP

            times = K_times[K]
            best_time = min(times[votation_method])
            best_k_for_times = times[votation_method].index(best_time) * K_STEP

            print(f'\t mejor accuracy: {best_accuracy} (k = {best_k_for_accuracy})')
            print(f'\t mejor f1: {best_f1} (k = {best_k_for_f1})')
            print(f'\t mejor tiempo: {best_time} (k = {best_k_for_times})')
            
            results_writer.writerow([
                K,
                votation_method,
                best_accuracy,
                best_k_for_accuracy,
                best_f1,
                best_k_for_f1,
                best_time,
                best_k_for_times
            ])
            
    for votation_method in votation_methods:
        best_accuracy_average = max(accuracys_average[votation_method].tolist())
        best_k_for_accuracy_average = accuracys_average[votation_method].tolist().index(best_accuracy_average) * K_STEP

        best_f1_average = max(f1s_average[votation_method].tolist())
        best_k_for_f1_average = f1s_average[votation_method].tolist().index(best_f1_average) * K_STEP

        best_time_average = min(times_average[votation_method].tolist())
        best_k_for_time_average = times_average[votation_method].tolist().index(best_time_average) * K_STEP

        results_writer.writerow([
            'average',
            votation_method,
            best_accuracy_average,
            best_k_for_accuracy_average,
            best_f1_average,
            best_k_for_f1_average,
            best_time_average,
            best_k_for_time_average
        ])


# Gráficos

# In[15]:


experimentPrefix = 'Experimento1'
xlabel = "Tamaño vecindario (k)"

if not os.path.exists(f'graficos/{experimentPrefix}'):
    os.makedirs(f'graficos/{experimentPrefix}') 
    
for votation_method in votation_methods:
    for K in range(KFOLDS):        
        # Gráfico 1: Accuracy
        accuracys = K_accuracys[K]
        accuracys_percentage = [a * 100 for a in accuracys[votation_method]]
        graficoTP2(f'{experimentPrefix}/fold={K}_votation-method={votation_method}_accuracy', ks, accuracys_percentage, xlabel, "Accuracy", colors[0])

        #Gráfico 2: F1 (media armónica entre presicion y recall)
        f1s = K_f1s[K]
        f1s_percentage = [a * 100 for a in f1s[votation_method]]
        graficoTP2(f'{experimentPrefix}/fold={K}_votation-method={votation_method}_media-armonica', ks, f1s_percentage, xlabel, "f1 = precision, recall", colors[1])

        # Gráfico 3: Tiempo de ejecución del algoritmo de predicción
        times = K_times[K]
        graficoTP2(f'{experimentPrefix}/fold={K}_votation-method={votation_method}_tiempo-de-ejecucion', ks, times[votation_method], xlabel, "Tiempo de ejecución (ms)", "green")

    
    graficoTP2(f'{experimentPrefix}/average_votation-method={votation_method}_accuracy', ks, accuracys_average[votation_method], xlabel, "Accuracy", colors[0])
    graficoTP2(f'{experimentPrefix}/average_votation-method={votation_method}_media-armonica', ks, f1s_average[votation_method], xlabel, "f1 = precision, recall", colors[1])
    graficoTP2(f'{experimentPrefix}/average_votation-method={votation_method}_tiempo-de-ejecucion', ks, times_average[votation_method], xlabel, "Tiempo de ejecución (ms)", "green")

